from django.conf import settings
from restosaur.contrib.django import API

api = API("api", force_script_name=settings.FORCE_SCRIPT_NAME)
root = api.resource("/")
some = api.resource("some")
subsome = api.resource("some/sub")


@root.get()
def root_view(ctx):
    return ctx.Response({"root": "ok"})


@some.get()
def some_view(ctx):
    return ctx.Response({"some": "ok"})


@subsome.get()
def subsome_view(ctx):
    return ctx.Response(
        {
            "some/sub": "ok",
            "root": ctx.url_for(root),
            "sub": ctx.url_for(subsome),
        }
    )


urlpatterns = api.urlpatterns()



from __future__ import absolute_import
from django.forms import *  # NOQA

import six


class NullBooleanSelect(NullBooleanSelect):
    def __init__(self, *args, **kw):
        super(NullBooleanSelect, self).__init__(*args, **kw)
        self.choices = (
                (u'', 'Unknown'),
                (u'1', 'Yes'),
                (u'0', 'No'))

    def value_from_datadict(self, data, files, name):
        value = data.get(name, None)
        if value is None or value == '':
            return None
        if value in ('True', 'true', '1', 1):
            return True
        return False


class NullBooleanField(NullBooleanField):
    widget = NullBooleanSelect


class BooleanField(BooleanField):
    widget = NullBooleanSelect

    def to_python(self, value):
        """Returns a Python boolean object."""
        if isinstance(
                value, six.string_types) and value.lower() in ('false', '0'):
            value = False
        else:
            if value is not None:
                value = bool(value)
        return value

    def validate(self, value):
        if value is None and self.required:
            raise ValidationError(self.error_messages['required'])


class ISODateField(DateTimeField):

    def strptime(self, value, format):
        import dateutil
        return dateutil.parser.parse(value)


class RestFormMixin:
    def _clean_fields(self):
        for name, field in self.fields.items():
            key = self.add_prefix(name)
            if not field.required and key not in self.data:
                continue

            value = field.widget.value_from_datadict(
                    self.data, self.files, key)
            try:
                if isinstance(field, FileField):
                    initial = self.initial.get(name, field.initial)
                    value = field.clean(value, initial)
                else:
                    value = field.clean(value)
                self.cleaned_data[name] = value
                if hasattr(self, 'clean_%s' % name):
                    value = getattr(self, 'clean_%s' % name)()
                    self.cleaned_data[name] = value
            except ValidationError as e:
                self._errors[name] = self.error_class(e.messages)
                if name in self.cleaned_data:
                    del self.cleaned_data[name]


class ModelForm(RestFormMixin, ModelForm):
    pass


class Form(RestFormMixin, Form):
    pass



from __future__ import absolute_import

import six

from django.template.loader import render_to_string


@six.python_2_unicode_compatible
class HTMLTemplate(object):
    def __init__(self, context, template_name, template_context):
        self.context = context
        self.template_name = template_name
        self.template_context = template_context
        self._content = None

    def __str__(self):
        if self._content is None:
            self._content = render_to_string(
                request=self.context.request, template_name=self.template_name,
                context=self.template_context)
        return self._content


def render_html(request_context, template_name, context=None):
    return HTMLTemplate(
            request_context, template_name=template_name,
            template_context=context)



import json
import datetime
import decimal
import six

from .datastructures import MultiValueDict


__all__ = [
        'JsonSerializer', 'MultiPartFormDataSerializer',
        'default_serializers']


class SerializeDeserializeError(Exception):
    pass


class DeserializationError(SerializeDeserializeError):
    pass


class SerializationError(SerializeDeserializeError):
    pass


_datetimes = (
    datetime.datetime, datetime.date, datetime.time)


class DefaultRestfulEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, _datetimes):
            return obj.isoformat()
        elif isinstance(obj, datetime.timedelta):
            return (datetime.datetime.min + obj).time().isoformat()
        elif isinstance(obj, decimal.Decimal):
            return float(str(obj))
        else:
            return super(DefaultRestfulEncoder, self).default(obj)


class DateTimeJsonSerializer(object):
    def dumps(self, obj):
        try:
            return json.dumps(obj, cls=DefaultRestfulEncoder)
        except (TypeError, ValueError) as ex:
            raise SerializationError(ex)

    def loads(self, txt):
        try:
            return json.loads(txt)
        except (TypeError, ValueError) as ex:
            raise DeserializationError(ex)


class JsonSerializer(object):
    def __init__(self):
        self._json = DateTimeJsonSerializer()

    def loads(self, ctx):
        if isinstance(ctx.raw, bytes):
            try:
                return self._json.loads(ctx.raw.decode(ctx.charset))
            except (TypeError, ValueError) as ex:
                raise DeserializationError(ex)
        else:
            try:
                return self._json.loads(ctx.raw)
            except (TypeError, ValueError) as ex:
                raise DeserializationError(ex)

    def dumps(self, ctx, data):
        try:
            return self._json.dumps(data)
        except (TypeError, ValueError) as ex:
            raise SerializationError(ex)


class MultiPartFormDataSerializer(object):
    def loads(self, ctx):
        data = MultiValueDict()
        data.update(ctx.data)
        data.update(ctx.files)
        return data

    def dumps(self, ctx, data):
        raise NotImplementedError


class HTMLSerializer(object):
    def loads(self, ctx):
        return ctx.raw

    def dumps(self, ctx, data):
        try:
            return six.text_type(data)
        except (TypeError, ValueError, UnicodeEncodeError,
                UnicodeDecodeError) as ex:
            raise SerializationError(ex)


class PlainTextSerializer(object):
    def loads(self, ctx):
        return ctx.raw

    def dumps(self, ctx, data):
        try:
            return six.text_type(data)
        except (TypeError, ValueError, UnicodeEncodeError,
                UnicodeDecodeError) as ex:
            raise SerializationError(ex)


class AlreadyRegistered(Exception):
    pass


class SerializersRegistry(object):
    def __init__(self):
        self._serializers = {}

    def _key(self, mimetype):
        return mimetype.lower()

    def __getitem__(self, item):
        return self._serializers[self._key(str(item))]

    def register(self, mimetype, instance):
        key = self._key(mimetype)
        if key in self._serializers:
            raise AlreadyRegistered(key)
        self._serializers[key] = instance

    def mimetypes(self):
        return self._serializers.keys()

    def items(self):
        return self._serializers.items()

    def contains(self, mimetype):
        return self._key(mimetype) in self._serializers


default_serializers = SerializersRegistry()
default_serializers.register(
        'application/json', JsonSerializer())
default_serializers.register(
        'text/html', HTMLSerializer())
default_serializers.register(
        'text/plain', PlainTextSerializer())
default_serializers.register(
        'multipart/form-data', MultiPartFormDataSerializer())


def register(mimetype, serializer):
    default_serializers.register(mimetype, serializer)


def get(mimetype):
    return default_serializers[mimetype]


def get_all():
    return default_serializers.items()



import unittest

from restosaur import API
from restosaur.contrib.django.dispatch import resource_dispatcher_factory


class BaseTestCase(unittest.TestCase):
    def setUp(self):
        super(BaseTestCase, self).setUp()

        from django.test import RequestFactory
        from django.http import StreamingHttpResponse, HttpResponse

        self.api = API('foo')
        self.resource = self.api.resource('test')
        self.stream_resource = self.api.resource('test-stream')

        self.StreamingHttpResponse = StreamingHttpResponse
        self.HttpResponse = HttpResponse

        @self.resource.get()
        def ctrl(ctx):
            return HttpResponse()

        @self.stream_resource.get()
        def ctrl_stream(ctx):
            return StreamingHttpResponse()

        self.rqfactory = RequestFactory()

    def call(self, resource, method, *args, **kw):
        rq = getattr(self.rqfactory, method)(resource.path, *args, **kw)
        return resource_dispatcher_factory(self.api, resource)(rq)


class PassingThroughNativeResponseTestCase(BaseTestCase):
    def test_http_response_class(self):
        resp = self.call(self.resource, 'get', HTTP_ACCEPT='application/json')
        self.assertTrue(isinstance(resp, self.HttpResponse))

    def test_http_streaming_response_class(self):
        resp = self.call(
            self.stream_resource, 'get', HTTP_ACCEPT='application/json')
        self.assertTrue(isinstance(resp, self.StreamingHttpResponse))



# coding: utf-8
from __future__ import unicode_literals

from restosaur import API
from restosaur.contrib.django.dispatch import resource_dispatcher_factory
from .utils import response_content_as_text

import unittest
import json


class MultiPartFormSerializerTestCase(unittest.TestCase):
    def setUp(self):
        super(MultiPartFormSerializerTestCase, self).setUp()

        from django.test import RequestFactory
        self.rqfactory = RequestFactory()

        self.api = API('foo')
        self.view = self.api.resource('/')

        @self.view.post('multipart/form-data')
        def view_post_json(ctx):
            return ctx.OK(ctx.body)

        @self.view.representation(media='application/json')
        def view_repr(obj, ctx):
            data = dict(obj.items())
            return data

    def call(self, resource, method, *args, **kw):
        rq = getattr(self.rqfactory, method)(resource.path, *args, **kw)
        return resource_dispatcher_factory(self.api, resource)(rq)

    def test_decoding_multipart_form(self):
        resp = self.call(
                self.view, 'post', data={'foo': 'bar'},
                HTTP_ACCEPT='application/json')
        self.assertEqual(resp.status_code, 200)
        data = json.loads(response_content_as_text(resp))
        self.assertEqual(data['foo'], 'bar')



"""
Restosaur - a tiny but real REST library

Author: Marcin Nowak <marcin.j.nowak@gmail.com>
"""

from __future__ import absolute_import

from . import resource  # NOQA
from . import responses  # NOQA

from .api import API  # NOQA


default_app_config = 'restosaur.apps.RestosaurAppConfig'



from django.conf import settings

AUTODISCOVER = getattr(settings, 'RESTOSAUR_AUTODISCOVER', True)
AUTODISCOVER_MODULE = getattr(
        settings, 'RESTOSAUR_AUTODISCOVER_MODULE', 'restapi')
AUTODISCOVER_MODULES = getattr(
        settings, 'RESTOSAUR_AUTODISCOVER_MODULES', [])



from .forms import Form


def filter_form_factory(model):
    fields = []

    for field in model._meta.fields:
        formfield = field.formfield()
        if formfield:
            formfield.required = False
            field_name = field.column if field.rel else field.name
            fields.append((field_name, formfield))

    fields = dict(fields)

    return type(Form)(model.__name__ + str('FilterForm'), (Form,), fields)


class QuerysetFilter(object):
    def __init__(self, queryset, form_class=None):
        self.form_class = form_class or filter_form_factory(queryset.model)
        self.queryset = queryset

    def narrow(self, data):
        form = self.form_class(data)
        if form.is_valid():
            data = form.cleaned_data
            really_cleaned_data = {}
            for key, value in data.items():
                if value is not None and value is not '' and value is not u'':
                    really_cleaned_data[key] = value
            return self.queryset.filter(**really_cleaned_data)
        return self.queryset



import unittest

from restosaur import API
from restosaur.dispatch import resource_dispatcher_factory


class CountMiddleware(object):
    def __init__(self):
        self.process_request_count = 0
        self.process_response_count = 0

    def process_request(self, *args, **kw):
        self.process_request_count += 1

    def process_response(self, *args, **kw):
        self.process_response_count += 1


class BreakRequestMiddleware(object):
    def process_request(self, request, context):
        return context.Created()


class BreakProcessingWithFalseMiddleware(object):
    def process_request(self, request, context):
        return False


class ReplaceResponseMiddleware(object):
    def process_response(self, request, response, context):
        return context.Created()


class ModifyResponseMiddleware(object):
    def process_response(self, request, response, context):
        response._status = 666


class ChangingContentMiddleware(object):
    def process_request(self, request, context):
        return context.OK('changed')


class BaseMiddlewareTestCase(unittest.TestCase):
    response_content = None
    default_content_type = 'text/plain'

    def setUp(self):
        self.api = self.createAPI()
        self.api.add_representation(
                str, 'text/plain', _transform_func=lambda x, ctx: x)

        self.resource = self.api.resource(
                '/', default_content_type=self.default_content_type)

        @self.resource.get()
        @self.resource.post()
        def resource_callback(ctx):
            return ctx.OK(self.response_content)

    def call(
            self, resource=None, method='get', parameters=None, data=None,
            headers=None):

        resource = resource or self.resource

        def response_builder(response, resource):
            return resource._http_response(response)

        def context_builder(api, resource, request):
            return api.make_context(**request)

        dispatcher = resource_dispatcher_factory(
                self.api, resource, response_builder, context_builder)

        return dispatcher({
            'path': resource.path,
            'method': method.upper(),
            'parameters': parameters,
            'data': data,
            'headers': headers,
            })


class NoMiddlewareTestCase(BaseMiddlewareTestCase):
    def createAPI(self):
        middlewares = []
        return API('/', middlewares=middlewares)

    def test_that_no_middlewares_creates_valid_response(self):
        resp = self.call()
        self.assertEqual(resp.status, 200)


class TwoMiddlewaresTestCase(BaseMiddlewareTestCase):
    def createAPI(self):
        self.countmiddleware = CountMiddleware()
        middlewares = [
                self.countmiddleware,
                self.countmiddleware
                ]
        return API('/', middlewares=middlewares)

    def test_middlewares_returnig_none_does_not_change_response(self):
        resp = self.call()
        self.assertEqual(resp.status, 200)

    def test_both_middlewares_processed_request(self):
        self.call()
        self.assertEqual(self.countmiddleware.process_request_count, 2)

    def test_both_middlewares_processed_response(self):
        self.call()
        self.assertEqual(self.countmiddleware.process_response_count, 2)


class RequestProcessingBreakTestCase(BaseMiddlewareTestCase):
    def createAPI(self):
        self.countmiddleware = CountMiddleware()
        middlewares = [
                self.countmiddleware,
                BreakRequestMiddleware(),
                self.countmiddleware
                ]
        return API('/', middlewares=middlewares)

    def test_middleware_changes_response(self):
        resp = self.call()
        self.assertEqual(resp.status, 201)

    def test_one_middleware_processed_request(self):
        self.call()
        self.assertEqual(self.countmiddleware.process_request_count, 1)

    def test_both_middlewares_processed_response(self):
        self.call()
        self.assertEqual(self.countmiddleware.process_response_count, 2)


class MiddlewareResponseReplacingTestCase(BaseMiddlewareTestCase):
    def createAPI(self):
        self.countmiddleware = CountMiddleware()
        middlewares = [
                self.countmiddleware,
                ReplaceResponseMiddleware(),
                self.countmiddleware
                ]
        return API('/', middlewares=middlewares)

    def test_middleware_changes_response(self):
        resp = self.call()
        self.assertEqual(resp.status, 201)

    def test_one_middleware_processed_request(self):
        self.call()
        self.assertEqual(self.countmiddleware.process_request_count, 2)

    def test_both_middlewares_processed_response(self):
        self.call()
        self.assertEqual(self.countmiddleware.process_response_count, 2)


class MiddlewareResponseModyfingTestCase(BaseMiddlewareTestCase):
    def createAPI(self):
        self.countmiddleware = CountMiddleware()
        middlewares = [
                self.countmiddleware,
                ModifyResponseMiddleware(),
                self.countmiddleware
                ]
        return API('/', middlewares=middlewares)

    def test_middleware_changes_response(self):
        resp = self.call()
        self.assertEqual(resp.status, 666)

    def test_one_middleware_processed_request(self):
        self.call()
        self.assertEqual(self.countmiddleware.process_request_count, 2)

    def test_both_middlewares_processed_response(self):
        self.call()
        self.assertEqual(self.countmiddleware.process_response_count, 2)


class AlwaysReturningResponseMiddlewareTestCase(BaseMiddlewareTestCase):
    def createAPI(self):
        self.countmiddleware = CountMiddleware()
        middlewares = [
                self.countmiddleware,
                BreakProcessingWithFalseMiddleware(),
                self.countmiddleware
                ]
        return API('/', middlewares=middlewares)

    def test_middleware_response_is_not_none(self):
        resp = self.call()
        self.assertIsNotNone(resp)

    def test_that_returning_false_does_not_break_request_processing(self):
        self.call()
        self.assertEqual(self.countmiddleware.process_request_count, 2)


class DeterminingResponseSerializerMiddlewareTestCase(BaseMiddlewareTestCase):
    resource_content = 'OK'

    def createAPI(self):
        self.countmiddleware = CountMiddleware()
        middlewares = [
                self.countmiddleware,
                ChangingContentMiddleware(),
                self.countmiddleware
                ]
        return API('/', middlewares=middlewares)

    def test_response_has_serializer_set(self):
        resp = self.call()
        self.assertTrue(hasattr(resp, 'serializer'))

    def test_response_has_content_set(self):
        resp = self.call()
        self.assertTrue(hasattr(resp, 'content'))

    def test_response_status_code_is_200OK(self):
        resp = self.call()
        self.assertEqual(resp.status, 200)

    def test_response_content_was_changed(self):
        resp = self.call()
        self.assertEqual(resp.content, 'changed')



from functools import cmp_to_key


def compare(a, b):
    return (a > b)-(a < b)


def parse_accept_header(accept):
    result = []
    for media_range in accept.split(","):
        parts = media_range.split(";")
        media_type = parts.pop(0).strip()
        # convert vendor-specific content types into something useful (see
        # docstring)
        typ, subtyp = media_type.split('/')
        # check for a + in the sub-type
        if '+' in subtyp:
            # if it exists, determine if the subtype is a vendor-specific type
            vnd, sep, extra = subtyp.partition('+')
            media_type = '{}/{}'.format(typ, extra)
        else:
            vnd = None
        q = 1.0
        for part in parts:
            (key, value) = part.lstrip().split("=", 1)
            key = key.strip()
            value = value.strip()
            if key == "q":
                q = float(value)
        result.append((media_type, vnd, q))
    result.sort(key=cmp_to_key(lambda x, y: -compare(x[2], y[2])))
    return result


def build_content_type_header(content_type, vendor):
    if vendor:
        type_, subtype = content_type.split('/')
        return '%s/%s+%s' % (type_, vendor, subtype)
    else:
        return content_type


def normalize_header_name(header):
    header = header.lower()
    if header.startswith('http_'):
        header = header.replace('http_', '', 1)
    return header.replace('_', '-')



from __future__ import absolute_import

try:
    from django.utils.encoding import force_text
except ImportError:
    from django.utils.encoding import force_str as force_text

from django.views import debug

from ...api import API as BaseAPI
from ...representations import ExceptionRepresentation
from ...urltemplate import RE_PARAMS
from .utils import autodiscover


def to_django_urlpattern(path):
    return RE_PARAMS.sub("/(?P<\\2>[^/]+)", path)


def django_html_exception(obj, ctx):
    if ctx.api.debug:
        resp = debug.technical_500_response(
            ctx.request,
            exc_type=obj.exc_type,
            exc_value=obj.exc_value,
            tb=obj.tb,
            status_code=obj.status_code,
        )
        return force_text(resp.content)
    return "<h1>Internal Server Error (%s)</h1>" % obj.status_code


class API(BaseAPI):
    def __init__(self, *args, **kw):
        from django.conf import settings

        charset = kw.pop("default_charset", None) or settings.DEFAULT_CHARSET
        debug = kw.pop("debug", settings.DEBUG)

        self.append_slash = kw.pop("append_slash", False)

        kw["default_charset"] = charset
        kw["debug"] = debug

        super(API, self).__init__(*args, **kw)

        self.add_representation(
            ExceptionRepresentation,
            content_type="text/html",
            _transform_func=django_html_exception,
            qvalue=0.2,
        )

    def get_urls(self):
        try:
            from django.urls import include
            from django.urls import re_path as url
        except ImportError:
            from django.conf.urls import include, url

        from django.views.decorators.csrf import csrf_exempt

        from .dispatch import resource_dispatcher_factory

        urls = []
        api_prefix = self.path.strip("/")

        for resource in self.resources:
            path = to_django_urlpattern(resource._path)
            if self.append_slash:
                path = path + "/"

            urls.append(
                url(
                    r"^%s$" % path,
                    csrf_exempt(resource_dispatcher_factory(self, resource)),
                )
            )

        if api_prefix:
            return [url(r"^%s/" % api_prefix, include(urls))]
        else:
            return urls

    def urlpatterns(self):
        return self.get_urls()

    def autodiscover(self, *args, **kw):
        """
        Shortcut for `restosaur.autodiscover()`
        """
        autodiscover(*args, **kw)



import warnings
from .contrib.django.forms import *  # NOQA

from .deprecation import RemovedInRestosaur08Warning

warnings.warn(
        '`%s` module is deprecated and will be removed in v0.8.'
        'Please replace your imports using '
        '``restosaur.contrib.django.forms`` module' % __name__,
        RemovedInRestosaur08Warning, stacklevel=2)



from importlib import import_module


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated
    by the last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError:
        msg = "%s doesn't look like a module path" % dotted_path
        raise ImportError(msg)

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError:
        msg = 'Module "%s" does not define a "%s" attribute/class' % (
            module_path, class_name)
        raise ImportError(msg)


def load_resource(string_path):
    return import_string(string_path)



import inspect
import six
from collections import defaultdict

from .loading import load_resource


DEFAULT_LINK_NAME = '__default__'


def model_class(instance_or_class):
    return type(instance_or_class) if not inspect.isclass(
            instance_or_class) else instance_or_class


class ModelAlreadyLinked(Exception):
    pass


class ModelNotLinked(Exception):
    pass


class MissingRequiredParameters(Exception):
    pass


class ModelLinksRegistry(object):
    def __init__(self):
        self._links = defaultdict(dict)

    def link(self, model, resource, name=None):
        """
        Create named link between a model and resource
        """

        name = name or DEFAULT_LINK_NAME
        cls = model_class(model)
        if name in self._links[cls]:
            raise ModelAlreadyLinked(
                    '%s is already linked to %s as "%s"' % (
                        cls, resource, name))
        self._links[cls][name] = resource

    def unlink(self, model, name=None):
        """
        Remove named link.
        If name is not provided, all model links will be removed.
        """

        cls = model_class(model)
        links = self._links[cls]

        if not links:
            raise ModelNotLinked(cls)

        if not name:
            self._links[cls] = {}
        else:
            if name not in links:
                raise ModelNotLinked(
                    '%s has no "%s" link registered' % (cls, name))
            del links[name]

    def instance_links(self, instance):
        """
        Return list of registered links for a model instance
        """

        return self._links[type(instance)]

    def class_links(self, cls):
        """
        Return list of registered links for a model class
        """

        return self._links[cls]

    def linked_resource(self, model, name=None):
        """
        Return resource for model class named link
        """
        name = name or DEFAULT_LINK_NAME

        if not inspect.isclass(model):
            model = type(model)

        try:
            resource = self._links[model][name]
        except KeyError:
            raise ModelNotLinked(
                    '%s has no "%s" link registered' % (model, name))

        if isinstance(resource, six.string_types):
            resource = load_resource(resource)
            self._links[model][name] = resource

        return resource

    def url(
            self, context, instance_or_class, name=None, parameters=None,
            query=None):
        if inspect.isclass(instance_or_class):
            return self.url_for_class(
                    context, instance_or_class, name=name, query=query,
                    parameters=parameters)
        else:
            return self.url_for_instance(
                    context, instance_or_class, name=name, query=query)

    def url_for_instance(
            self, context, instance, name=None, query=None):
        """
        Generate URL for a model instance based on registered links.
        Path parameters will be read from the model instance.
        Optional query args may be provided.
        """

        name = name or DEFAULT_LINK_NAME
        resource = self.linked_resource(type(instance), name=name)

        if not resource._required_parameters:
            return resource.uri(self, query=query)

        params = {}

        for parameter in resource._required_parameters:
            try:
                params[parameter] = getattr(instance, parameter)
            except AttributeError:
                try:
                    params[parameter] = instance[parameter]
                except (KeyError, TypeError, ValueError):
                    raise ValueError(
                        'Can\'t construct URL parameter "%s" from `%s`' % (
                            parameter, instance))

        return resource.uri(context, params=params, query=query)

    def url_for_class(
            self, context, cls, name=None, parameters=None, query=None):
        """
        Generate URL for a model class based on registered links.
        Path parameters should be provided.
        Optional query args may be also provided.
        """

        name = name or DEFAULT_LINK_NAME
        resource = self.linked_resource(cls, name=name)
        parameters = parameters or {}

        missing_parameters = []

        for parameter in resource._required_parameters:
            if parameter not in parameters:
                missing_parameters.append(parameter)

        if missing_parameters:
            raise MissingRequiredParameters(missing_parameters)

        return resource.uri(context, params=parameters, query=query)



import os

from setuptools import find_packages, setup

README_PATH = os.path.join(os.path.dirname(__file__), "README.md")

setup(
    name="restosaur",
    version="0.7.0b21",
    description="Framework independent RESTful library",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 1.11",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Developers",
    ],
    author="Marcin Nowak",
    author_email="marcin.j.nowak@gmail.com",
    url="https://github.com/marcinn/restosaur",
    install_requires=["times2==0.9", "six"],
    keywords="web rest python django",
    packages=find_packages("."),
    include_package_data=True,
    test_suite="nose.collector",
    zip_safe=True,
    long_description=open(README_PATH).read(),
    long_description_content_type="text/markdown",
)



import cgi

__version__ = '1.5.2'
__author__ = 'Joe Gregorio'
__email__ = 'joe@bitworking.org'
__license__ = 'MIT License'
__credits__ = ''


class MimeTypeParseException(ValueError):
    pass


def parse_mime_type(mime_type):
    """Parses a mime-type into its component parts.

    Carves up a mime-type and returns a tuple of the (type, subtype, params)
    where 'params' is a dictionary of all the parameters for the media range.
    For example, the media range 'application/xhtml;q=0.5' would get parsed
    into:

       ('application', 'xhtml', {'q', '0.5'})

    :rtype: (str,str,dict)
    """
    full_type, params = cgi.parse_header(mime_type)
    # Java URLConnection class sends an Accept header that includes a
    # single '*'. Turn it into a legal wildcard.
    if full_type == '*':
        full_type = '*/*'

    type_parts = full_type.split('/') if '/' in full_type else None
    if not type_parts or len(type_parts) > 2:
        raise MimeTypeParseException(
            "Can't parse type \"{}\"".format(full_type))

    (type, subtype) = type_parts

    return (type.strip(), subtype.strip(), params)


def parse_media_range(range):
    """Parse a media-range into its component parts.

    Carves up a media range and returns a tuple of the (type, subtype,
    params) where 'params' is a dictionary of all the parameters for the media
    range.  For example, the media range 'application/*;q=0.5' would get parsed
    into:

       ('application', '*', {'q', '0.5'})

    In addition this function also guarantees that there is a value for 'q'
    in the params dictionary, filling it in with a proper default if
    necessary.

    :rtype: (str,str,dict)
    """
    (type, subtype, params) = parse_mime_type(range)
    params.setdefault('q', params.pop('Q', None))  # q is case insensitive
    try:
        if not params['q'] or not 0 <= float(params['q']) <= 1:
            params['q'] = '1'
    except ValueError:  # from float()
        params['q'] = '1'

    return (type, subtype, params)


def quality_and_fitness_parsed(mime_type, parsed_ranges):
    """Find the best match for a mime-type amongst parsed media-ranges.

    Find the best match for a given mime-type against a list of media_ranges
    that have already been parsed by parse_media_range(). Returns a tuple of
    the fitness value and the value of the 'q' quality parameter of the best
    match, or (-1, 0) if no match was found. Just as for quality_parsed(),
    'parsed_ranges' must be a list of parsed media ranges.

    :rtype: (float,int)
    """
    best_fitness = -1
    best_fit_q = 0
    (target_type, target_subtype, target_params) = \
        parse_media_range(mime_type)

    for (type, subtype, params) in parsed_ranges:

        # check if the type and the subtype match
        type_match = (
            type in (target_type, '*') or
            target_type == '*'
        )
        subtype_match = (
            subtype in (target_subtype, '*') or
            target_subtype == '*'
        )

        # if they do, assess the "fitness" of this mime_type
        if type_match and subtype_match:

            # 100 points if the type matches w/o a wildcard
            fitness = type == target_type and 100 or 0

            # 10 points if the subtype matches w/o a wildcard
            fitness += subtype == target_subtype and 10 or 0

            # 1 bonus point for each matching param besides "q"
            param_matches = sum([
                1 for (key, value) in target_params.items()
                if key != 'q' and key in params and value == params[key]
            ])
            fitness += param_matches

            # finally, add the target's "q" param (between 0 and 1)
            fitness += float(target_params.get('q', 1))

            if fitness > best_fitness:
                best_fitness = fitness
                best_fit_q = params['q']

    return float(best_fit_q), best_fitness


def quality_parsed(mime_type, parsed_ranges):
    """Find the best match for a mime-type amongst parsed media-ranges.

    Find the best match for a given mime-type against a list of media_ranges
    that have already been parsed by parse_media_range(). Returns the 'q'
    quality parameter of the best match, 0 if no match was found. This function
    behaves the same as quality() except that 'parsed_ranges' must be a list of
    parsed media ranges.

    :rtype: float
    """

    return quality_and_fitness_parsed(mime_type, parsed_ranges)[0]


def quality(mime_type, ranges):
    """Return the quality ('q') of a mime-type against a list of media-ranges.

    Returns the quality 'q' of a mime-type when compared against the
    media-ranges in ranges. For example:

    >>> quality('text/html','text/*;q=0.3, text/html;q=0.7,
                  text/html;level=1, text/html;level=2;q=0.4, */*;q=0.5')
    0.7

    :rtype: float
    """
    parsed_ranges = [parse_media_range(r) for r in ranges.split(',')]

    return quality_parsed(mime_type, parsed_ranges)


def best_match(supported, header):
    """Return mime-type with the highest quality ('q') from list of candidates.

    Takes a list of supported mime-types and finds the best match for all the
    media-ranges listed in header. The value of header must be a string that
    conforms to the format of the HTTP Accept: header. The value of 'supported'
    is a list of mime-types. The list of supported mime-types should be sorted
    in order of increasing desirability, in case of a situation where there is
    a tie.

    >>> best_match(['application/xbel+xml', 'text/xml'],
                   'text/*;q=0.5,*/*; q=0.1')
    'text/xml'

    :rtype: str
    """
    split_header = _filter_blank(header.split(','))
    parsed_header = [parse_media_range(r) for r in split_header]
    weighted_matches = []
    pos = 0
    for mime_type in supported:
        weighted_matches.append((
            quality_and_fitness_parsed(mime_type, parsed_header),
            pos,
            mime_type
        ))
        pos += 1
    weighted_matches.sort()

    return weighted_matches[-1][0][0] and weighted_matches[-1][2] or ''


def _filter_blank(i):
    """Return all non-empty items in the list."""
    for s in i:
        if s.strip():
            yield s



from restosaur import dispatch as restosaur_dispatch
from restosaur.context import QueryDict
from restosaur.exceptions import Http404 as RestosaurHttp404
from restosaur.headers import normalize_header_name

from django.http import Http404, HttpResponse
from django.http.response import HttpResponseBase


def build_context(api, resource, request):
    try:
        # Django may raise RawPostDataException sometimes;
        # i.e. when processing POST multipart/form-data;
        # In that cases we can't access raw body anymore, sorry

        raw_body = request.body
    except Exception:
        raw_body = None

    parameters = {}

    if request.resolver_match:
        parameters.update(request.resolver_match.kwargs)

    parameters.update(QueryDict(list(request.GET.lists())))

    headers = dict(
        map(
            lambda x: (normalize_header_name(x[0]), x[1]),
            filter(lambda x: x[0].startswith("HTTP_"), request.META.items()),
        )
    )

    try:
        content_length = int(request.META["CONTENT_LENGTH"])
    except (KeyError, TypeError, ValueError):
        content_length = 0

    content_type = request.META.get("CONTENT_TYPE")

    return api.make_context(
        host=request.get_host(),
        path=request.path,
        method=request.method,
        parameters=parameters,
        data=request.POST,
        files=request.FILES,
        raw=raw_body,
        charset=request.encoding or api.default_charset,
        secure=request.is_secure(),
        encoding=request.GET.encoding,
        resource=resource,
        request=request,
        headers=headers,
        content_type=content_type,
        content_length=content_length,
    )


def http_response_builder(response, resource):
    if response is None:
        return HttpResponse()

    if isinstance(response, HttpResponseBase):
        return response

    response = resource._http_response(response)

    if response.serializer:
        content = response.serializer.dumps(response.context, response.content)
    else:
        content = ""

    httpresp = HttpResponse(content, status=response.status)

    if response.content_type:
        httpresp["Content-Type"] = response.content_type

    for header, value in response.headers.items():
        httpresp[header] = value

    return httpresp


class DjangoMethodDispatcher(restosaur_dispatch.DefaultResourceDispatcher):
    def do_call(self, callback, ctx, args=None, kwargs=None):
        try:
            return super(DjangoMethodDispatcher, self).do_call(
                callback, ctx, args=args, kwargs=kwargs
            )
        except Http404 as ex:
            raise RestosaurHttp404(ex)


def resource_dispatcher_factory(api, resource):
    return restosaur_dispatch.resource_dispatcher_factory(
        api,
        resource,
        http_response_builder,
        build_context,
        dispatcher_class=DjangoMethodDispatcher,
    )
